/*
 * Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once 

#ifndef TENSORFLOW_COMMON_OP_H
#define TENSORFLOW_COMMON_OP_H

#include "fastertransformer/common.h"
#include "fastertransformer/tf_op/tf_traits.h"

#include "tensorflow/core/framework/op.h"
#include "tensorflow/core/framework/op_kernel.h"
#include "tensorflow/core/framework/shape_inference.h"
#include "tensorflow/core/framework/register_types.h"
#include "tensorflow/core/lib/core/errors.h"
#include "tensorflow/core/framework/types.h"
#include "tensorflow/core/framework/tensor_types.h"
#include "third_party/eigen3/unsupported/Eigen/CXX11/Tensor"

#include <cuda_fp16.h>

namespace tensorflow
{
namespace
{
using CPUDevice = Eigen::ThreadPoolDevice;
using GPUDevice = Eigen::GpuDevice;

template <typename T>
class CommonOp : public OpKernel
{
public:
  explicit CommonOp(OpKernelConstruction *context) : OpKernel(context) {
    try
    {
      check_cuda_error(cublasCreate(&cublas_handle_));
      check_cuda_error(cublasLtCreate(&cublaslt_handle_));
    }
    catch(std::runtime_error& error)
    {
    OP_REQUIRES(context, false, errors::Internal(error.what()));
    }
  };

  template<typename DataType_>
  void get_tensor(OpKernelContext *context, int tensor_id, const DataType_** tensor_ptr, int off_set = 0){
    *tensor_ptr = reinterpret_cast<const DataType_ *>(context->input(tensor_id).flat<T>().data()) + off_set;
    OP_REQUIRES(context, *tensor_ptr != nullptr, errors::InvalidArgument("tensor %d is null", tensor_id));
  }

  cublasHandle_t get_cublas_handler() {return cublas_handle_; }
  cublasLtHandle_t get_cublaslt_handler() {return cublaslt_handle_;}

  ~CommonOp() { cublasDestroy(cublas_handle_); cublasLtDestroy(cublaslt_handle_);}
private:
  cublasHandle_t cublas_handle_;
  cublasLtHandle_t cublaslt_handle_;

};

} //namespace
} //namespace tensorflow

#endif
